#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


#include "memoryalloc.h"
#include "hashtable.h"
#include "hashes.h"
#include "letlittolithash.h"
#include "stringops.h"
#include "vector.h"
#include "state.h"
#include "stateset.h"
#include "ast.h"

struct AST *ASTNew(char **input, char **output, size_t length)
{
  struct AST *result = (struct AST *)
    SafeMalloc(sizeof (struct AST));
  ASTMake(result, input, output, length);
  return result;
}

void ASTDelete(struct AST *ast)
{
  if (ast)
  {
    ASTDestroy(ast);
    free(ast);
  }
}

static size_t GetCommonPrefixLength(char *a, size_t aLen, char *b, size_t bLen)
{
  size_t len = aLen < bLen ? aLen : bLen;
  size_t i;

  for (i = 0; i < len; i++)
  {
    if (a[i] != b[i])
    {
      return i;
    }
  }
  return len;
}

static void CheckInputWords(char *currentWord, size_t currentWordLen)
{
  if (currentWordLen >= MAX_STRING)
  {
    FatalError("Input too long: %s\n", currentWord);
  }
  if (currentWordLen < 1)
  {
    FatalError("Input too short, empty word not allowed!");
  }
}

static void CheckWords(char *currentWord, size_t currentWordLen, char *currentOutput)
{
  CheckInputWords(currentWord, currentWordLen);
  if (strlen(currentOutput) >= MAX_STRING)
  {
    FatalError("Output too long: %s\n", currentOutput);
  }
}

void ASTMake(struct AST *ast, char **input, char **output, size_t length)
{
  char *prevWord, *currentWord, *currentOutput;
  size_t i, j, prefixLength, prevWordLen, currentWordLen;

  for (i = 0; i < MEMORY_BLOCK; ++i)
  {
    StateMake(&ast->tempStates[i], 1);
  }

  StateSetMake(&ast->minStates, ast->tempStates);

  ast->initialState = NULL;

  prevWord = empty;
  prevWordLen = 0;
  StateSetClear(&ast->tempStates[0]);

  for (j = 0; j < length; ++j)
  {
    currentWord = input[j];
    currentOutput = output[j];
    currentWordLen = strlen(currentWord);
    CheckWords(currentWord, currentWordLen, currentOutput);
    /* calculate the length of the longest common prefix */
    prefixLength = GetCommonPrefixLength(
      currentWord, currentWordLen, prevWord, prevWordLen);
    /* minimize the states from the suffix of the previous word */
    for (i = prevWordLen; i > prefixLength; --i)
    {
      StateSetTransition(&ast->tempStates[i - 1],
        prevWord[i - 1],
        StateSetFindMinimized(&ast->minStates, &ast->tempStates[i]));
    }

    ASTMakeTail(ast, currentWord, currentWordLen, currentOutput, prefixLength);

    /* set previous word */
    prevWord = currentWord;
    prevWordLen = currentWordLen;
  }
  ASTMinimize(ast, prevWord, prevWordLen);
}

void ASTDestroy(struct AST *ast)
{
  size_t i;

  if (ast)
  {
    for (i = 0; i < MEMORY_BLOCK; ++i)
    {
      StateDestroy(&ast->tempStates[i], 0);
    }
    StateSetDestroy(&ast->minStates);
  }
}

void ASTAddPair(struct AST *ast, char *currentWord, char *currentOutput)
{
  struct State *currentState, *nextState;
  size_t currentWordLen = strlen(currentWord);
  size_t i = 0;
  size_t lastIndex = 0;
  char c;

  CheckWords(currentWord, currentWordLen, currentOutput);

  currentState = StateSetClone(&ast->minStates, NULL, '_', ast->initialState, &ast->tempStates[0]);

  /* find the prefix: */
  for (;;)
  {
    currentState = &ast->tempStates[i];
    c = currentWord[i];
    nextState = StateGetTransition(currentState, c);
    if (nextState == NULL)
    {
      break;
    }
    nextState = StateSetClone(&ast->minStates, currentState, c, nextState, &ast->tempStates[i + 1]);
    lastIndex = i + 1;
    if (i < currentWordLen - 1)
    {
      ++i;
    }
    else
    {
      break;
    }
  }

  StateSetDePropagate(&ast->minStates, currentWord, lastIndex);

  if (nextState == NULL)
  {
    ASTMakeTail(ast, currentWord, currentWordLen, currentOutput, i);
  }
  else
  {
    ASTMakeTail(ast, currentWord, currentWordLen, currentOutput, i + 1);
  }
  ASTMinimize(ast, currentWord, currentWordLen);
}

void ASTRemovePair(struct AST *ast, char *currentWord)
{
  struct State *currentState, *nextState;
  size_t currentWordLen = strlen(currentWord);
  size_t i = 0;
  size_t lastIndex = 0;
  char c;
  unsigned char hasNext;

  CheckInputWords(currentWord, currentWordLen);

  currentState = StateSetClone(&ast->minStates, NULL, '_', ast->initialState, &ast->tempStates[0]);

  /* find the prefix: */
  for (;;)
  {
    currentState = &ast->tempStates[i];
    c = currentWord[i];
    nextState = StateGetTransition(currentState, c);
    if (nextState == NULL)
    {
      break;
    }
    nextState = StateSetClone(&ast->minStates, currentState, c, nextState, &ast->tempStates[i + 1]);
    lastIndex = i + 1;
    if (i < currentWordLen - 1)
    {
      ++i;
    }
    else
    {
      break;
    }
  }

  if (
    (lastIndex < 1) ||
    (lastIndex != currentWordLen) ||
    (!StateIsFinal(&ast->tempStates[lastIndex])))
  {
    FatalError("This word is not present: %s\n", currentWord);
  }

  StateSetFinal(&ast->tempStates[lastIndex], 0);

  while (lastIndex > 0)
  {
    if (StateIsFinal(&ast->tempStates[lastIndex]))
    {
      break;
    }
    hasNext = 0;
    for (c = FIRST_LETTER; c <= LAST_LETTER; ++c)
    {
      if (StateGetTransition(&ast->tempStates[lastIndex], (char)c) != NULL)
      {
        hasNext = 1;
        break;
      }
    }
    if (hasNext)
    {
      break;
    }
    --lastIndex;
    StateSetTransition(
      &ast->tempStates[lastIndex], currentWord[lastIndex], NULL);
  }

  StateSetDePropagate(&ast->minStates, currentWord, lastIndex);

  ASTMinimize(ast, currentWord, lastIndex);
}

unsigned ASTIsKeyPresent(struct AST *ast, char *word)
{
  struct State * state;
  size_t length;
  size_t i;

  state = ast->initialState;
  length = strlen(word);
  for (i = 0; i < length; ++i)
  {
    if (state == NULL)
    {
      break;
    }
    state = StateGetTransition(state, word[i]);
  }
  return i == length && state != NULL && StateIsFinal(state);
}

void ASTMakeTail(struct AST *ast, char *currentWord, size_t currentWordLen, char *currentOutputIn, size_t prefixLength)
{
  char commonPrefix[MEMORY_BLOCK];
  char wordSuffix[MEMORY_BLOCK];
  char currentOutput[MEMORY_BLOCK];
  char newOutput[MEMORY_BLOCK];
  char *out;
  size_t i, j;
  size_t commonPrefixLength;
  size_t c;

  strcpy(currentOutput, currentOutputIn);

  /* initialize the tail states of the current word */
  for (i = prefixLength + 1; i <= currentWordLen; ++i)
  {
    StateSetClear(&ast->tempStates[i]);
    StateSetTransition(&ast->tempStates[i - 1],
      currentWord[i - 1],
      &ast->tempStates[i]);
  }
  /* propagate output */
  for (j = 1; j <= prefixLength; ++j)
  {
    out = StateGetOutput(&ast->tempStates[j - 1], currentWord[j - 1]);
    commonPrefixLength = GetCommonPrefix(out, currentOutput, commonPrefix);
    GetMinusSized(commonPrefixLength, out, wordSuffix);
    StateSetOutput(&ast->tempStates[j - 1], currentWord[j - 1], commonPrefix);
    for (c = FIRST_LETTER; c <= LAST_LETTER; ++c)
    {
      if (StateGetTransition(&ast->tempStates[j], (char)c) != NULL)
      {
        Concatenate(wordSuffix, StateGetOutput(&ast->tempStates[j], (char)c), newOutput);
        StateSetOutput(&ast->tempStates[j], (char)c, newOutput);
      }
    }
    if (StateIsFinal(&ast->tempStates[j]))
    {
      Concatenate(wordSuffix, StateGetStateOutput(&ast->tempStates[j]), newOutput);
      StateSetStateOutput(&ast->tempStates[j], newOutput);
    }
    GetMinusSized(commonPrefixLength, currentOutput, currentOutput);
  }
  if (prefixLength < currentWordLen)
  {
    StateSetOutput(&ast->tempStates[prefixLength],
      currentWord[prefixLength], currentOutput);
    *currentOutput = '\0';
  }
  /* set final state */
  StateSetFinal(&ast->tempStates[currentWordLen], 1);
  StateSetStateOutput(&ast->tempStates[currentWordLen], currentOutput);
}

void ASTMinimize(struct AST *ast, char *prevWord, size_t prevWordLen)
{
  size_t i;
  for (i = prevWordLen; i >= 1; --i) {
    StateSetTransition(&ast->tempStates[i - 1], prevWord[i - 1],
      StateSetFindMinimized(&ast->minStates, &ast->tempStates[i]));
  }
  ast->initialState = StateSetFindMinimized(&ast->minStates, &ast->tempStates[0]);
}

struct State *ASTGetInitialState(struct AST *ast)
{
  return ast->initialState;
}

static void GraphVizStateDump(FILE * file, struct State * state, unsigned initial)
{
  char * shape;
  char * psy;
  unsigned final;
  int id;

  psy = StateGetStateOutput(state);
  id = StateGetID(state);
  final = StateIsFinal(state);
  if (initial)
  {
    if (final)
    {
      shape = "doubleellipse";
    }
    else
    {
      shape = "ellipse";
    }
  }
  else
  {
    if (final)
    {
      shape = "doublecircle";
    }
    else
    {
      shape = "circle";
    }
  }
  fprintf(file, "  node [shape = %s]; ", shape);
  if ((psy != NULL) && (*psy))
  {
    fprintf(file, "q%d [label = \"q%d(\\\"%s\\\")\"];\n", id, id, psy);
  }
  else
  {
    fprintf(file, "q%d;\n", id);
  }
}

static void GraphVizStateLinksDump(FILE * file, struct State * state)
{
  struct State *delta;
  char *lambda;
  size_t i;
  int id;

  id = StateGetID(state);

  for (i = FIRST_LETTER; i <= LAST_LETTER; ++i)
  {
    delta = StateGetTransition(state, (char)i);
    if (delta != NULL)
    {
      lambda = StateGetOutput(state, (char)i);
      /* q1 -> q2 [label = "'a' | \"37\""]; */
      fprintf(file, "  q%d -> q%d [label = \"'%c'", id, StateGetID(delta), (char)i);
      if (*lambda)
      {
        fprintf(file, " | \\\"%s\\\"", lambda);
      }
      fprintf(file, "\"];\n");
    }
  }
}

void ASTPrint(struct AST *ast, FILE *file, unsigned normal)
{
  struct Hashtable *h = &ast->minStates.states;
  struct List *node;
  struct State *state;
  size_t i;

  if (normal)
  {
    if (ast->initialState != NULL)
    {
      fprintf(file, "initial is : %d\n", ast->initialState->id);
    }
    else
    {
      fprintf(file, "initial is NULL!\n");
    }
    for (i = 0; i < h->capacity; ++i)
    {
      for (node = h->data[i]; node != NULL; node = node->next)
      {
        if (node->key != NULL)
        {
          state = (struct State *)node->key;
          StatePrint(state, file);
        }
      }
    }
  }
  else
  {
    fprintf(file, "digraph AcyclicSequenceTransducer {\n");
    fprintf(file, "\n");
    fprintf(file, "  rankdir = LR;\n");
    /*
      //fprintf(file, "  size=\"8,5\";\n");
    */
    fprintf(file, "  orientation=land;\n");
    fprintf(file, "\n");

    if (h->size > 1)
    {

      if (ast->initialState != NULL)
      {
        GraphVizStateDump(file, ast->initialState, 1);
      }
      for (i = 0; i < h->capacity; ++i)
      {
        for (node = h->data[i]; node != NULL; node = node->next)
        {
          if (node->key != NULL)
          {
            state = (struct State *)node->key;
            if (state != ast->initialState)
            {
              GraphVizStateDump(file, state, 0);
            }
          }
        }
      }

      fprintf(file, "\n");

      if (ast->initialState != NULL)
      {
        GraphVizStateLinksDump(file, ast->initialState);
      }
      for (i = 0; i < h->capacity; ++i)
      {
        for (node = h->data[i]; node != NULL; node = node->next)
        {
          if (node->key != NULL)
          {
            state = (struct State *)node->key;
            if (state != ast->initialState)
            {
              GraphVizStateLinksDump(file, state);
            }
          }
        }
      }

      fprintf(file, "\n");

    }

    fprintf(file, "}\n");
  }
}
