#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>


#include "memoryalloc.h"
#include "hashtable.h"
#include "hashes.h"
#include "letlittolithash.h"
#include "stringops.h"
#include "vector.h"
#include "state.h"

/*
  //static int cc = 0;
  //static int dc = 0;
*/

/**
 * This hashtable type is
 * intended to hold permanent state->itself
 * links. it automatically deletes the
 * permanent state once it is removed.
 */
struct HashtableType
StateToStateTableType = {
  (HashType)StateHashCode,
  LiteralCopy,
  LiteralCopy,
  (DeleteKeyType)StateTerminate,
  (DeleteDataType)LiteralDelete,
  (EqualKeysType)StateEquals,
  NULL
};

static int staticId = 0;
static struct Hashtable *pDeltas = NULL;
static struct Hashtable *pLambdas = NULL;

void StateEndStateChange(struct State *state)
{
  state->hashCalculated = 0;
}

void StateInitTables()
{
  if (pDeltas != NULL)
  {
    FatalError("StateInitTables() called twice!\n");
  }
  pDeltas = HashtableNew(&LetLitToLitTableType);
  pLambdas = HashtableNew(&LetLitToStringTableType);
}

void StateDestroyTables()
{
  if (pDeltas == NULL)
  {
    FatalError("StateDestroyTables() called before StateInitTables()!\n");
  }
  HashtableDelete(pDeltas);
  HashtableDelete(pLambdas);
}

struct State *StateNew(unsigned isTemp)
{
  struct State *state = (struct State *)SafeMalloc(sizeof (struct State));
  StateMake(state, isTemp);
  return state;
}

void StateDelete(struct State *state)
{
  if (state)
  {
    StateDestroy(state, 1);
    free(state);
  }
}

void StateTerminate(struct State *state)
{
  if (state)
  {
    StateDestroy(state, 0);
    free(state);
  }
}

void StateMake(struct State *state, unsigned isTemp)
{
  if (isTemp)
  {
    size_t i;

    state->stateData.tempStateData = (struct TempStateData *)
      SafeMalloc(sizeof (struct TempStateData));
    state->stateData.tempStateData->psy[0] = '\0';
    for (i = 0; i < LETTERS_COUNT; ++i)
    {
      state->stateData.tempStateData->lambda[i][0] = '\0';
      state->stateData.tempStateData->delta[i] = NULL;
    }
  }
  else
  {
    state->stateData.permanentStateData.psy = NULL;
  }
  state->backLinks = 0;
  state->hashCode = 0;
  state->id = staticId++;
  state->copyOf = -1;
  state->hashCalculated = (unsigned char)0;
  state->isFinal = (unsigned char)0;
  state->isTemp = (unsigned char)isTemp;
  state->forceModifyChildrenOnDestroy = (unsigned char)0;
}

void StateDestroyInternal(struct State *state, unsigned freeTempStateData, unsigned modifyChildren)
{
  if (state->isTemp)
  {
    size_t i;
    struct State *to;

    if (modifyChildren || state->forceModifyChildrenOnDestroy)
    {
      for (i = 0; i < LETTERS_COUNT; ++i)
      {
        to = state->stateData.tempStateData->delta[i];
        if (to != NULL)
        {
          --to->backLinks;
        }
      }
    }
    if (freeTempStateData)
    {
      free(state->stateData.tempStateData);
    }
  }
  else
  {
    struct LetLit letlit;
    struct State *to;
    letlit.literal = state;
    for (letlit.letter = FIRST_LETTER; letlit.letter <= LAST_LETTER; ++letlit.letter)
    {
      to = (struct State *)HashtableGet(pDeltas, &letlit);
      if (to != NULL)
      {
        HashtableRemove(pDeltas, &letlit);
        if (modifyChildren || state->forceModifyChildrenOnDestroy)
        {
          --to->backLinks;
        }
      }
      HashtableRemove(pLambdas, &letlit);
    }
    if (state->stateData.permanentStateData.psy != NULL)
    {
      free(state->stateData.permanentStateData.psy);
    }
  }
}

void StateDestroy(struct State *state, unsigned modifyChildren)
{
  StateDestroyInternal(state, 1, modifyChildren);
}

unsigned StateIsFinal(struct State *state)
{
  return state->isFinal;
}

void StateSetFinal(struct State *state, unsigned isFinal)
{
  state->isFinal = (unsigned char)isFinal;
  if (!isFinal)
  {
    if (state->isTemp)
    {
      state->stateData.tempStateData->psy[0] = '\0';
    }
    else
    {
      if (state->stateData.permanentStateData.psy != NULL)
      {
        free(state->stateData.permanentStateData.psy);
        state->stateData.permanentStateData.psy = NULL;
      }
    }
  }
  StateEndStateChange(state);
}

struct State *StateGetTransition(struct State *state, char letter)
{
  if (state->isTemp)
  {
    return state->stateData.tempStateData->delta[CHAR_VAR_TO_INDEX(letter)];
  }
  else
  {
    struct LetLit letlit;
    letlit.letter = letter;
    letlit.literal = state;
    return (struct State *)HashtableGet(pDeltas, &letlit);
  }
}

void StateSetTransition(struct State *state, char letter, struct State *to)
{
  if (state->isTemp)
  {
    struct State *old;
    old = state->stateData.tempStateData->delta[CHAR_VAR_TO_INDEX(letter)];
    state->stateData.tempStateData->delta[CHAR_VAR_TO_INDEX(letter)] = to;
    if (old != NULL)
    {
      if (to != NULL && old->id == to->id)
      {
        /* no need to re-add! */
        return;
      }
      --old->backLinks;
    }
    if (to != NULL)
    {
      ++to->backLinks;
    }
  }
  else
  {
    struct LetLit letlit;
    struct State *old;

    letlit.letter = letter;
    letlit.literal = state;
    old = (struct State *)HashtableGet(pDeltas, &letlit);
    if (old != NULL)
    {
      if (to != NULL && old->id == to->id)
      {
        /* no need to re-add! */
        return;
      }
      HashtableRemove(pDeltas, &letlit);
      --old->backLinks;
    }
    if (to != NULL)
    {
      HashtablePut(pDeltas, &letlit, to);
      ++to->backLinks;
    }
  }

  StateEndStateChange(state);
}

int StateGetID(struct State *state)
{
  return state->id;
}

char *StateGetStateOutput(struct State *state)
{
  if (!(state->isFinal))
  {
    return empty;
  }
  if (state->isTemp)
  {
    return state->stateData.tempStateData->psy;
  }
  else
  {
    if (state->stateData.permanentStateData.psy == NULL)
    {
      return empty;
    }
    return state->stateData.permanentStateData.psy;
  }
}

void StateSetStateOutput(struct State *state, char *psy)
{
  if (state->isTemp)
  {
    if (psy == NULL)
    {
      psy = empty;
    }
    memmove(state->stateData.tempStateData->psy, psy, strlen(psy) + 1);
  }
  else
  {
    if (state->stateData.permanentStateData.psy)
    {
      free(state->stateData.permanentStateData.psy);
    }

    if (psy == NULL || psy[0] == '\0')
    {
      state->stateData.permanentStateData.psy = NULL;
    }
    else
    {
      state->stateData.permanentStateData.psy =
        AllocateAnotherString(psy);
      memmove(state->stateData.permanentStateData.psy, psy, strlen(psy) + 1);
    }
  }

  StateEndStateChange(state);
}

char *StateGetOutput(struct State *state, char letter)
{
  if (state->isTemp)
  {
    return state->stateData.tempStateData->lambda[CHAR_VAR_TO_INDEX(letter)];
  }
  else
  {
    struct LetLit letlit;
    char *result;

    letlit.letter = letter;
    letlit.literal = state;
    result = (char *)HashtableGet(pLambdas, &letlit);
    if (result == NULL)
    {
      result = empty;
    }
    return result;
  }
}

void StateSetOutput(struct State *state, char letter, char *lambda)
{
  if (state->isTemp)
  {
    if (lambda == NULL)
    {
      lambda = empty;
    }
    memmove(
      state->stateData.tempStateData->lambda[CHAR_VAR_TO_INDEX(letter)],
      lambda, strlen(lambda) + 1);
  }
  else
  {
    struct LetLit letlit;

    letlit.letter = letter;
    letlit.literal = state;
    if ((lambda != NULL) && (*lambda == '\0'))
    {
      lambda = NULL;
    }
    HashtableRemove(pLambdas, &letlit);
    if (lambda != NULL)
    {
      HashtablePut(pLambdas, &letlit, lambda);
    }
  }

  StateEndStateChange(state);
}

unsigned StateHasBackLinks(struct State *state)
{
  return state->backLinks > 0;
}

void StateSetClear(struct State *state)
{
  StateDestroyInternal(state, 0, 1);

  if (state->isTemp)
  {
    size_t i;

    state->stateData.tempStateData->psy[0] = '\0';
    for (i = 0; i < LETTERS_COUNT; ++i)
    {
      state->stateData.tempStateData->lambda[i][0] = '\0';
      state->stateData.tempStateData->delta[i] = NULL;
    }
  }
  else
  {
    state->stateData.permanentStateData.psy = NULL;
  }
  state->backLinks = 0;
  state->hashCode = 0;
  /*
   * Do Not modify the id!
   * state->id = staticId++;
   */
  state->copyOf = -1;
  state->hashCalculated = (unsigned char)0;
  state->isFinal = (unsigned char)0;
  /*
   * Do Not modify isTemp!
   * state->isTemp = isTemp;
   */

  StateEndStateChange(state);
}

void StateCopyInto(struct State *state, struct State *other)
{
  size_t i;
  struct State *delta;

  StateSetClear(other);

  other->hashCode = state->hashCode;
  other->copyOf = state->id;
  other->isFinal = state->isFinal;

  StateSetStateOutput(other, StateGetStateOutput(state));

  for (i = FIRST_LETTER; i <= LAST_LETTER; ++i)
  {
    delta = StateGetTransition(state, (char)i);
    if (delta != NULL)
    {
      StateSetTransition(other, (char)i, delta);
      StateSetOutput(other, (char)i, StateGetOutput(state, (char)i));
    }
  }

  other->hashCalculated = state->hashCalculated;
}

struct State *StateGetCopy(struct State *state, unsigned isTemp)
{
  struct State *other;
  other = StateNew(isTemp);
  StateCopyInto(state, other);
  return other;
}

void StateGetCommonPrefix(struct State *state, char prefix[MEMORY_BLOCK])
{
  size_t i = 0;
  char *psy;
  char *lambda;
  unsigned first = 1;

  *prefix = '\0';
  if (state->isFinal)
  {
    psy = StateGetStateOutput(state);

    if (psy == NULL || (!(*psy)))
    {
      /**
       * No need to continue!
       */
      return;
    }
    first = 0;
    memmove(prefix, psy, strlen(psy) + 1);
  }
  for (i = FIRST_LETTER; i <= LAST_LETTER; ++i)
  {
    if (StateGetTransition(state, (char)i) != NULL)
    {
      lambda = StateGetOutput(state, (char)i);
      if (first)
      {
        first = 0;
        memmove(prefix, lambda, strlen(lambda) + 1);
      }
      else
      {
        GetCommonPrefix(prefix, lambda, prefix);
      }
    }
  }

}

void StateRemoveCommonPrefix(struct State *state, size_t size)
{
  char other[MEMORY_BLOCK];
  size_t i;
  char *lambda;
  char *psy;

  if (size == 0)
  {
    return;
  }

  if (state->isFinal)
  {
    psy = StateGetStateOutput(state);
    if (psy != NULL && (*psy))
    {
      GetMinusSized(size, psy, psy);
      if (!state->isTemp)
      {
        if (*psy)
        {
          strcpy(other, psy);
          free(psy);
          psy = AllocateAnotherString(other);
          strcpy(psy, other);
          state->stateData.permanentStateData.psy = psy;
        }
        else
        {
          free(psy);
          state->stateData.permanentStateData.psy = NULL;
        }
      }
    }
  }
  for (i = FIRST_LETTER; i <= LAST_LETTER; ++i)
  {
    if (StateGetTransition(state, (char)i) != NULL)
    {
      lambda = StateGetOutput(state, (char)i);
      if (lambda != NULL && (*lambda))
      {
        GetMinusSized(size, lambda, other);
        StateSetOutput(state, (char)i, other);
      }
    }
  }
  StateEndStateChange(state);
}

void StateCalculateHash(struct State *state)
{
  struct State *delta;
  char *lambda;
  char *psy;
  size_t i;
  size_t result;

  result = state->isFinal ? 11 : 17;
  for (i = FIRST_LETTER; i <= LAST_LETTER; ++i)
  {
    delta = StateGetTransition(state, (char)i);
    if (delta != NULL)
    {
      lambda = StateGetOutput(state, (char)i);
      result = result * 11 + i;
      result = result * 11 + delta->id;
      result = result * 11 + StringHash(lambda);
    }
  }
  psy = StateGetStateOutput(state);
  if (psy != NULL && (*psy))
  {
    result = result * 11 + StringHash(psy);
  }
  state->hashCode = result;
  state->hashCalculated = 1;
}

size_t StateHashCode(struct State *state)
{
  if (!state->hashCalculated)
  {
    StateCalculateHash(state);
  }
  return state->hashCode;
}

unsigned StateEquals(struct State *state, struct State *other)
{
  struct State *delta;
  char *lambda;
  size_t i;

  if (state == other)
  {
    return 1;
  }
  if (state == NULL || other == NULL)
  {
    return 0;
  }
  if (StateHashCode(state) != StateHashCode(other))
  {
    return 0;
  }
  if (state->isFinal != other->isFinal)
  {
    return 0;
  }
  if (state->isFinal)
  {
    if (!StringEqual(StateGetStateOutput(state), StateGetStateOutput(other)))
    {
      return 0;
    }
  }
  for (i = FIRST_LETTER; i <= LAST_LETTER; ++i)
  {
    delta = StateGetTransition(state, (char)i);
    if (StateGetTransition(other, (char)i) != delta)
    {
      return 0;
    }
    if (delta != NULL)
    {
      lambda = StateGetOutput(state, (char)i);
      if (!StringEqual(lambda, StateGetOutput(other, (char)i)))
      {
        return 0;
      }
    }
  }
  return 1;
}

void StatePrint(struct State *state, FILE *file)
{
  struct State *delta;
  char *lambda;
  char *psy;
  size_t i;
  unsigned prev;

  fprintf(file, "id = %d; ", state->id);
  fprintf(file, "copy of = %d; ", state->copyOf);
  fprintf(file, "delta = {");
  prev = 0;
  for (i = FIRST_LETTER; i <= LAST_LETTER; ++i)
  {
    delta = StateGetTransition(state, (char)i);
    if (delta != NULL)
    {
      if (prev)
      {
        fprintf(file, ", ");
      }
      fprintf(file, "%c=%d", (char)i, delta->id);
      prev = 1;
    }
  }
  fprintf(file, "}; ");
  fprintf(file, "lambda = {");
  prev = 0;
  for (i = FIRST_LETTER; i <= LAST_LETTER; ++i)
  {
    lambda = StateGetOutput(state, (char)i);
    if (lambda != NULL && (*lambda))
    {
      if (prev)
      {
        fprintf(file, ", ");
      }
      fprintf(file, "%c=%s", (char)i, lambda);
      prev = 1;
    }
  }
  fprintf(file, "}; ");
  psy = StateGetStateOutput(state);
  fprintf(file, "psy = %s; ", psy != NULL ? psy : empty);
  if (state->isFinal)
  {
    fprintf(file, "f = true; ");
  }
  else
  {
    fprintf(file, "f = false; ");
  }
  fprintf(file, "back = %d\n", state->backLinks);
}

